import json
from inspect import isawaitable
from typing import Any, Callable, Dict, List, Optional, Type, Union, cast

from graphql import (
    ExecutionContext,
    GraphQLError,
    GraphQLSchema,
    MiddlewareManager,
)

from .constants import (
    CONTENT_TYPE_JSON,
    CONTENT_TYPE_TEXT_HTML,
    CONTENT_TYPE_TEXT_PLAIN,
    DATA_TYPE_JSON,
    DATA_TYPE_MULTIPART,
    HTTP_STATUS_200_OK,
    HTTP_STATUS_400_BAD_REQUEST,
    HTTP_STATUS_405_METHOD_NOT_ALLOWED,
)
from .exceptions import HttpBadRequestError, HttpError
from .explorer import Explorer, ExplorerGraphiQL
from .file_uploads import combine_multipart_data
from .format_error import format_error
from .graphql import graphql_sync
from .types import (
    ContextValue,
    ErrorFormatter,
    ExtensionList,
    GraphQLResult,
    MiddlewareList,
    QueryParser,
    RootValue,
    ValidationRules,
)
from .utils import context_value_one_arg_deprecated

try:
    from multipart import parse_form
except ImportError:

    def parse_form(*_args, **_kwargs):
        raise NotImplementedError(
            "WSGI file uploads requires 'python-multipart' library."
        )


__all__ = ["FormData", "GraphQL", "GraphQLMiddleware"]

Extensions = Union[
    Callable[[Any, Optional[ContextValue]], ExtensionList], ExtensionList
]

Middlewares = Union[
    Callable[[Any, Optional[ContextValue]], MiddlewareList], MiddlewareList
]


class GraphQL:
    """WSGI application implementing the GraphQL server."""

    def __init__(
        self,
        schema: GraphQLSchema,
        *,
        context_value: Optional[ContextValue] = None,
        root_value: Optional[RootValue] = None,
        query_parser: Optional[QueryParser] = None,
        validation_rules: Optional[ValidationRules] = None,
        debug: bool = False,
        introspection: bool = True,
        explorer: Optional[Explorer] = None,
        logger: Optional[str] = None,
        error_formatter: ErrorFormatter = format_error,
        extensions: Optional[Extensions] = None,
        middleware: Optional[Middlewares] = None,
        middleware_manager_class: Optional[Type[MiddlewareManager]] = None,
        execution_context_class: Optional[Type[ExecutionContext]] = None,
    ) -> None:
        """Initializes the WSGI app.

        # Required arguments

        `schema`: an instance of GraphQL schema to execute queries against.

        # Optional arguments

        `context_value`: a `ContextValue` to use by this server for context.
        Defaults to `{"request": request}` dictionary where `request` is
        an WSGI environment dictionary.

        `root_value`: a `RootValue` to use by this server for root value.
        Defaults to `None`.

        `query_parser`: a `QueryParser` to use by this server. Defaults to
        `graphql.parse`.

        `validation_rules`: a `ValidationRules` list or callable returning a
        list of extra validation rules server should use to validate the
        GraphQL queries. Defaults to `None`.

        `debug`: a `bool` controlling in server should run in debug mode or
        not. Controls details included in error data returned to clients.
        Defaults to `False`.

        `introspection`: a `bool` controlling if server should allow the
        GraphQL introspection queries. If `False`, introspection queries will
        fail to pass the validation. Defaults to `True`.

        `explorer`: an instance of `Explorer` subclass to use when the server
        receives an HTTP GET request. If not set, default GraphQL explorer
        for your version of Ariadne is used.

        `logger`: a `str` with name of logger or logger instance server
        instance should use for logging errors. If not set, a logger named
        `ariadne` is used.

        `error_formatter`: an `ErrorFormatter` this server should use to format
        GraphQL errors returned to clients. If not set, default formatter
        implemented by Ariadne is used.

        `extensions`: an `Extensions` list or callable returning a
        list of extensions server should use during query execution. Defaults
        to no extensions.

        `middleware`: a `Middlewares` list or callable returning a list of
        middlewares server should use during query execution. Defaults to no
        middlewares.

        `middleware_manager_class`: a `MiddlewareManager` type or subclass to
        use for combining provided middlewares into single wrapper for resolvers
        by the server. Defaults to `graphql.MiddlewareManager`. Is only used
        if `extensions` or `middleware` options are set.

        `execution_context_class`: custom `ExecutionContext` type to use by
        this server to execute the GraphQL queries. Defaults to standard
        context type implemented by the `graphql`.
        """

        self.context_value = context_value
        self.root_value = root_value
        self.query_parser = query_parser
        self.validation_rules = validation_rules
        self.debug = debug
        self.introspection = introspection
        self.logger = logger
        self.error_formatter = error_formatter
        self.extensions = extensions
        self.middleware = middleware
        self.middleware_manager_class = middleware_manager_class or MiddlewareManager
        self.execution_context_class = execution_context_class
        self.schema = schema

        if explorer:
            self.explorer = explorer
        else:
            self.explorer = ExplorerGraphiQL()

    def __call__(self, environ: dict, start_response: Callable) -> List[bytes]:
        """An entrypoint to the WSGI application.

        Returns list of bytes with response body.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `start_response`: a callable used to begin new HTTP response.

        Details about the arguments and their usage are described in PEP 3333:

        https://peps.python.org/pep-3333/
        """
        try:
            return self.handle_request(environ, start_response)
        except GraphQLError as error:
            return self.handle_graphql_error(error, start_response)
        except HttpError as error:
            return self.handle_http_error(error, start_response)

    def handle_graphql_error(
        self, error: GraphQLError, start_response: Callable
    ) -> List[bytes]:
        """Handles a `GraphQLError` raised from `handle_request` and returns an
        error response to the client.

        Returns list of bytes with response body.

        # Required arguments

        `error`: a `GraphQLError` instance.

        `start_response`: a callable used to begin new HTTP response.
        """
        start_response(
            HTTP_STATUS_400_BAD_REQUEST, [("Content-Type", CONTENT_TYPE_JSON)]
        )
        error_json = {"errors": [{"message": error.message}]}
        return [json.dumps(error_json).encode("utf-8")]

    def handle_http_error(
        self, error: HttpError, start_response: Callable
    ) -> List[bytes]:
        """Handles a `HttpError` raised from `handle_request` and returns an
        error response to the client.

        Returns list of bytes with response body.

        # Required arguments

        `error`: a `HttpError` instance.

        `start_response`: a callable used to begin new HTTP response.
        """
        start_response(error.status, [("Content-Type", CONTENT_TYPE_TEXT_PLAIN)])
        response_body = error.message or error.status
        return [str(response_body).encode("utf-8")]

    def handle_request(self, environ: dict, start_response: Callable) -> List[bytes]:
        """Handles WSGI HTTP request and returns a a response to the client.

        Returns list of bytes with response body.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `start_response`: a callable used to begin new HTTP response.
        """
        if environ["REQUEST_METHOD"] == "GET" and self.introspection:
            return self.handle_get(environ, start_response)
        if environ["REQUEST_METHOD"] == "POST":
            return self.handle_post(environ, start_response)

        return self.handle_not_allowed_method(environ, start_response)

    def handle_get(self, environ: dict, start_response) -> List[bytes]:
        """Handles WSGI HTTP GET request and returns a a response to the client.

        Returns list of bytes with response body.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `start_response`: a callable used to begin new HTTP response.
        """
        explorer_html = self.explorer.html(environ)
        if isawaitable(explorer_html):
            raise ValueError("Explorer HTML can't be awaitable.")
        if not explorer_html:
            return self.handle_not_allowed_method(environ, start_response)

        start_response(HTTP_STATUS_200_OK, [("Content-Type", CONTENT_TYPE_TEXT_HTML)])
        return [cast(str, explorer_html).encode("utf-8")]

    def handle_post(self, environ: dict, start_response: Callable) -> List[bytes]:
        """Handles WSGI HTTP POST request and returns a a response to the client.

        Returns list of bytes with response body.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `start_response`: a callable used to begin new HTTP response.
        """
        data = self.get_request_data(environ)
        result = self.execute_query(environ, data)
        return self.return_response_from_result(start_response, result)

    def get_request_data(self, environ: dict) -> dict:
        """Extracts GraphQL request data from request.

        Returns a `dict` with GraphQL query data that was not yet validated.

        # Required arguments

        `environ`: a WSGI environment dictionary.
        """
        content_type = environ.get("CONTENT_TYPE", "")
        content_type = content_type.split(";")[0]

        if content_type == DATA_TYPE_JSON:
            return self.extract_data_from_json_request(environ)
        if content_type == DATA_TYPE_MULTIPART:
            return self.extract_data_from_multipart_request(environ)

        raise HttpBadRequestError(
            "Posted content must be of type {} or {}".format(
                DATA_TYPE_JSON, DATA_TYPE_MULTIPART
            )
        )

    def extract_data_from_json_request(self, environ: dict) -> Any:
        """Extracts GraphQL data from JSON request.

        Returns a `dict` with GraphQL query data that was not yet validated.

        # Required arguments

        `environ`: a WSGI environment dictionary.
        """
        request_content_length = self.get_request_content_length(environ)
        request_body = self.get_request_body(environ, request_content_length)

        try:
            return json.loads(request_body)
        except ValueError as ex:
            raise HttpBadRequestError("Request body is not a valid JSON") from ex

    def get_request_content_length(self, environ: dict) -> int:
        """Validates and returns value from `Content-length` header.

        Returns an `int` with content length.

        Raises a `HttpBadRequestError` error if `Content-length` header is
        missing or invalid.

        # Required arguments

        `environ`: a WSGI environment dictionary.
        """
        try:
            content_length = int(environ.get("CONTENT_LENGTH", 0))
            if content_length < 1:
                raise HttpBadRequestError(
                    "Content length header is missing or incorrect"
                )
            return content_length
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError(
                "Content length header is missing or incorrect"
            ) from ex

    def get_request_body(self, environ: dict, content_length: int) -> bytes:
        """Returns request's body.

        Returns `bytes` with request body of specified length.

        Raises a `HttpBadRequestError` error if request body is empty.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `content_length`: an `int` with content length.
        """
        if not environ.get("wsgi.input"):
            raise HttpBadRequestError("Request body cannot be empty")
        request_body = environ["wsgi.input"].read(content_length)
        if not request_body:
            raise HttpBadRequestError("Request body cannot be empty")
        return request_body

    def extract_data_from_multipart_request(self, environ: dict) -> Any:
        """Extracts GraphQL data from `multipart/form-data` request.

        Returns an unvalidated `dict` with GraphQL query data.

        # Required arguments

        `environ`: a WSGI environment dictionary.
        """
        try:
            form = parse_multipart_request(environ)
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError("Malformed request data") from ex

        try:
            operations = json.loads(form.getvalue("operations"))
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError(
                "Request 'operations' multipart field is not a valid JSON"
            ) from ex
        try:
            files_map = json.loads(form.getvalue("map"))
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError(
                "Request 'map' multipart field is not a valid JSON"
            ) from ex

        return combine_multipart_data(operations, files_map, form.files)

    def execute_query(self, environ: dict, data: dict) -> GraphQLResult:
        """Executes GraphQL query and returns its result.

        Returns a `GraphQLResult`, a two items long `tuple` with `bool` for
        success and JSON-serializable `data` to return to client.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `data`: a GraphQL data.
        """
        context_value = self.get_context_for_request(environ, data)
        extensions = self.get_extensions_for_request(environ, context_value)
        middleware = self.get_middleware_for_request(environ, context_value)

        return graphql_sync(
            self.schema,
            data,
            context_value=context_value,
            root_value=self.root_value,
            query_parser=self.query_parser,
            validation_rules=self.validation_rules,
            debug=self.debug,
            introspection=self.introspection,
            logger=self.logger,
            error_formatter=self.error_formatter,
            extensions=extensions,
            middleware=middleware,
            middleware_manager_class=self.middleware_manager_class,
            execution_context_class=self.execution_context_class,
        )

    def get_context_for_request(
        self, environ: dict, data: dict
    ) -> Optional[ContextValue]:
        """Returns GraphQL context value for HTTP request.

        Default `ContextValue` for WSGI application is a `dict` with single
        `request` key that contains WSGI environment dictionary.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `data`: a GraphQL data.
        """
        if callable(self.context_value):
            try:
                return self.context_value(environ, data)  # type: ignore
            except TypeError:  # TODO: remove in 0.20
                context_value_one_arg_deprecated()
                return self.context_value(environ)  # type: ignore
        return self.context_value or {"request": environ}

    def get_extensions_for_request(
        self, environ: dict, context: Optional[ContextValue]
    ) -> ExtensionList:
        """Returns extensions to use when handling the GraphQL request.

        Returns `ExtensionList`, a list of extensions to use or `None`.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `context`: a `ContextValue` for this request.
        """
        if callable(self.extensions):
            return self.extensions(environ, context)
        return self.extensions

    def get_middleware_for_request(
        self, environ: dict, context: Optional[ContextValue]
    ) -> Optional[MiddlewareList]:
        """Returns GraphQL middlewares to use when handling the GraphQL request.

        Returns `MiddlewareList`, a list of middlewares to use or `None`.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `context`: a `ContextValue` for this request.
        """
        middleware = self.middleware
        if callable(middleware):
            middleware = middleware(environ, context)
        if middleware:
            return cast(MiddlewareList, middleware)
        return None

    def return_response_from_result(
        self, start_response: Callable, result: GraphQLResult
    ) -> List[bytes]:
        """Returns WSGI response from GraphQL result.

        Returns a list of bytes with response body.

        # Required arguments

        `start_response`: a WSGI callable that initiates new response.

        `result`: a `GraphQLResult` for this request.
        """
        success, response = result
        status_str = HTTP_STATUS_200_OK if success else HTTP_STATUS_400_BAD_REQUEST
        start_response(status_str, [("Content-Type", CONTENT_TYPE_JSON)])
        return [json.dumps(response).encode("utf-8")]

    def handle_not_allowed_method(
        self, environ: dict, start_response: Callable
    ) -> List[bytes]:
        """Handles request for unsupported HTTP method.

        Returns 200 response for `OPTIONS` request and 405 response for other
        methods. All responses have empty body.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `start_response`: a WSGI callable that initiates new response.
        """
        allowed_methods = ["OPTIONS", "POST"]
        if self.introspection:
            allowed_methods.append("GET")

        if environ["REQUEST_METHOD"] == "OPTIONS":
            status_str = HTTP_STATUS_200_OK
        else:
            status_str = HTTP_STATUS_405_METHOD_NOT_ALLOWED

        headers = [
            ("Content-Type", CONTENT_TYPE_TEXT_PLAIN),
            ("Content-Length", 0),
            ("Allow", ", ".join(allowed_methods)),
        ]

        start_response(status_str, headers)
        return []


class GraphQLMiddleware:
    """Simple WSGI middleware routing requests to either app or GraphQL."""

    def __init__(
        self, app: Callable, graphql_app: Callable, path: str = "/graphql/"
    ) -> None:
        """Initializes the WSGI middleware.

        Returns response from either application or GraphQL application

        # Required arguments

        `app`: a WSGI application to route the request to if its path doesn't
        match the `path` option.

        `graphql_app`: a WSGI application to route the request to if its path
        matches the `path` option.

        # Optional arguments

        `path`: a `str` with a path to the GraphQL application. Defaults to
        `/graphql/`.
        """
        self.app = app
        self.path = path
        self.graphql_app = graphql_app

        if not callable(app):
            raise TypeError("app must be a callable WSGI application")

        if not path:
            raise ValueError("path can't be empty")

        if path == "/":
            raise ValueError(
                "WSGI middleware can't use root path together with "
                "application callable"
            )

    def __call__(self, environ: dict, start_response: Callable) -> List[bytes]:
        """An entrypoint to the WSGI middleware.

        Returns list of bytes with response body.

        # Required arguments

        `environ`: a WSGI environment dictionary.

        `start_response`: a callable used to start new HTTP response.
        """
        if not environ["PATH_INFO"].startswith(self.path):
            return self.app(environ, start_response)
        return self.graphql_app(environ, start_response)


def parse_multipart_request(environ: dict) -> "FormData":
    content_type = environ.get("CONTENT_TYPE")
    headers = {"Content-Type": content_type}
    form_data = FormData(content_type)

    parse_form(
        headers,
        environ["wsgi.input"],
        form_data.on_field,
        form_data.on_file,
    )

    return form_data


class FormData:
    """Feature-limited alternative of deprecated `cgi` standard package.

    Holds the data from `multipart/form-data` request.

    # Attributes

    `charset`: a string with charset extracted from `Content-type` header.
    Defaults to `latin-1`.

    `fields`: an `dict` with form's fields names and values.

    `files`: an `dict` with form's files names and values.
    """

    charset: str
    fields: Dict[str, Any]
    files: Dict[str, Any]

    def __init__(self, content_type: Optional[str]):
        """Initializes form data instance.

        # Optional arguments

        `content_type`: a string with content type header's value. If not
        provided, `latin-1` is used for content encoding.
        """
        self.encoding = self.parse_charset(content_type) or "latin-1"
        self.fields = {}
        self.files = {}

    def parse_charset(self, content_type: Optional[str]) -> Optional[str]:
        """Parses charset from `Content-type` header

        Returns none if `content_type` is not provided, empty or missing the
        `charset=` declaration.

        # Optional arguments

        `content_type`: a string with content type header's value.
        """
        if not content_type:
            return None

        if "charset=" not in content_type:
            return None

        charset = content_type[content_type.index("charset=") + 8 :].strip()
        if ";" in charset:
            charset = charset[: charset.index(";")].strip()

        return charset.lower() or None

    def on_field(self, field):
        """Callback for HTTP request parser to provide field data.

        Field name and value is decoded using the encoding stored in `encoding`
        attribute and stored in `fields` attribute.
        """
        field_name = field.field_name.decode(self.encoding)
        field_value = field.value.decode(self.encoding)
        self.fields[field_name] = field_value

    def on_file(self, file):
        """Callback for HTTP request parser to provide file data.

        File's field name is decoded using the encoding stored in `encoding`
        attribute and stored in `files` attribute.
        """
        field_name = file.field_name.decode(self.encoding)
        self.files[field_name] = file

    def getvalue(self, field_name: str) -> str:
        """Get form field's value.

        Returns field's value or empty string if field didn't exist.

        # Required arguments

        `field_name`: a `str` with name of form field to return the value for.
        """
        return self.fields.get(field_name, "")



from .default_query import escape_default_query
from .explorer import Explorer
from .template import read_template, render_template

APOLLO_HTML = read_template("apollo_sandbox.html")

DEFAULT_QUERY = "# Write your query or mutation here"


class ExplorerApollo(Explorer):
    def __init__(
        self,
        title: str = "Ariadne GraphQL",
        default_query: str = DEFAULT_QUERY,
    ) -> None:
        self.parsed_html = render_template(
            APOLLO_HTML,
            {
                "title": title,
                "default_query": escape_default_query(default_query),
            },
        )

    def html(self, _):
        return self.parsed_html



import os
from functools import partial
from inspect import isawaitable
from typing import Any, Callable, Dict, Optional, Union

from graphql import GraphQLResolveInfo
from opentracing import Scope, Tracer, global_tracer
from opentracing.ext import tags
from starlette.datastructures import UploadFile

from ...types import ContextValue, Extension, Resolver
from .utils import format_path, should_trace

try:
    from multipart.multipart import File
except ImportError:

    class File:  # type: ignore
        """Mock upload file used when python-multipart is not installed."""


ArgFilter = Callable[[Dict[str, Any], GraphQLResolveInfo], Dict[str, Any]]


class OpenTracingExtension(Extension):
    _arg_filter: Optional[ArgFilter]
    _root_scope: Scope
    _tracer: Tracer

    def __init__(self, *, arg_filter: Optional[ArgFilter] = None) -> None:
        self._arg_filter = arg_filter
        self._tracer = global_tracer()

    def request_started(self, context: ContextValue):
        self._root_scope = self._tracer.start_active_span("GraphQL Query")
        self._root_scope.span.set_tag(tags.COMPONENT, "graphql")

    def request_finished(self, context: ContextValue):
        self._root_scope.close()

    async def resolve(
        self, next_: Resolver, obj: Any, info: GraphQLResolveInfo, **kwargs
    ):
        if not should_trace(info):
            result = next_(obj, info, **kwargs)
            if isawaitable(result):
                result = await result
            return result

        with self._tracer.start_active_span(info.field_name) as scope:
            span = scope.span
            span.set_tag(tags.COMPONENT, "graphql")
            span.set_tag("graphql.parentType", info.parent_type.name)

            graphql_path = ".".join(
                map(str, format_path(info.path))  # pylint: disable=bad-builtin
            )
            span.set_tag("graphql.path", graphql_path)

            if kwargs:
                filtered_kwargs = self.filter_resolver_args(kwargs, info)
                for kwarg, value in filtered_kwargs.items():
                    span.set_tag(f"graphql.param.{kwarg}", value)

            result = next_(obj, info, **kwargs)
            if isawaitable(result):
                result = await result
            return result

    def filter_resolver_args(
        self, args: Dict[str, Any], info: GraphQLResolveInfo
    ) -> Dict[str, Any]:
        args_to_trace = copy_args_for_tracing(args)

        if not self._arg_filter:
            return args_to_trace

        return self._arg_filter(args_to_trace, info)


class OpenTracingExtensionSync(OpenTracingExtension):
    def resolve(
        self, next_: Resolver, obj: Any, info: GraphQLResolveInfo, **kwargs
    ):  # pylint: disable=invalid-overridden-method
        if not should_trace(info):
            result = next_(obj, info, **kwargs)
            return result

        with self._tracer.start_active_span(info.field_name) as scope:
            span = scope.span
            span.set_tag(tags.COMPONENT, "graphql")
            span.set_tag("graphql.parentType", info.parent_type.name)

            graphql_path = ".".join(
                map(str, format_path(info.path))  # pylint: disable=bad-builtin
            )
            span.set_tag("graphql.path", graphql_path)

            if kwargs:
                filtered_kwargs = self.filter_resolver_args(kwargs, info)
                for kwarg, value in filtered_kwargs.items():
                    span.set_tag(f"graphql.param.{kwarg}", value)

            result = next_(obj, info, **kwargs)
            return result


def opentracing_extension(*, arg_filter: Optional[ArgFilter] = None):
    return partial(OpenTracingExtension, arg_filter=arg_filter)


def opentracing_extension_sync(*, arg_filter: Optional[ArgFilter] = None):
    return partial(OpenTracingExtensionSync, arg_filter=arg_filter)


def copy_args_for_tracing(value: Any) -> Any:
    if isinstance(value, dict):
        return {k: copy_args_for_tracing(v) for k, v in value.items()}
    if isinstance(value, list):
        return [copy_args_for_tracing(v) for v in value]
    if isinstance(value, (UploadFile, File)):
        return repr_upload_file(value)
    return value


def repr_upload_file(upload_file: Union[UploadFile, File]) -> str:
    if isinstance(upload_file, File):
        filename = upload_file.file_name
    else:
        filename = upload_file.filename

    mime_type: Union[str, None]

    if isinstance(upload_file, File):
        mime_type = "not/available"
    else:
        mime_type = upload_file.content_type

    if isinstance(upload_file, File):
        size = upload_file.size
    else:
        file_ = upload_file.file
        file_.seek(0, os.SEEK_END)
        size = file_.tell()
        file_.seek(0)

    return (
        f"{type(upload_file)}(mime_type={mime_type}, size={size}, filename={filename})"
    )



from typing import Callable, Optional, Tuple

from graphql import (
    GraphQLField,
    GraphQLInputObjectType,
    GraphQLObjectType,
    GraphQLSchema,
)

from .resolvers import resolve_to
from .utils import convert_camel_case_to_snake

"""
A type of a function implementing a strategy for names conversion in schema. 
Passed as an option to `make_executable_schema` and `convert_schema_names` 
functions.

Takes three arguments:

`name`: a `str` with schema name to convert.

`schema`: the GraphQL schema in which names are converted.

`path`: a tuple of `str` representing a path to the schema item which name 
is being converted.

Returns a string with the Python name to use.
"""
SchemaNameConverter = Callable[[str, GraphQLSchema, Tuple[str, ...]], str]

GRAPHQL_SPEC_TYPES = (
    "__Directive",
    "__EnumValue",
    "__Field",
    "__InputValue",
    "__Schema",
    "__Type",
)


def convert_schema_names(
    schema: GraphQLSchema,
    name_converter: Optional[SchemaNameConverter],
) -> None:
    """Set mappings in GraphQL schema from `camelCase` names to `snake_case`.

    This function scans GraphQL schema and:

    If objects field has name in `camelCase` and this field doesn't have a
    resolver already set on it, new resolver is assigned to it that resolves
    it's value from object attribute or dict key named like `snake_case`
    version of field's name.

    If object's field has argument in `camelCase` and this argument doesn't have
    the `out_name` attribute already set, this attribute is populated with
    argument's name converted to `snake_case`

    If input's field has name in `camelCase` and it's `out_name` attribute is
    not already set, this attribute is populated with field's name converted
    to `snake_case`.

    Schema is mutated in place.

    Generally you shouldn't call this function yourself, as its part of
    `make_executable_schema` logic, but its part of public API for other
    libraries to use.

    # Required arguments

    `schema`: a GraphQL schema to update.

    `name_converter`: an `SchemaNameConverter` function to use to convert the
    names from `camelCase` to `snake_case`. If not provided, default one
    based on `convert_camel_case_to_snake` is used.
    """
    name_converter = name_converter or default_schema_name_converter

    for type_name, graphql_type in schema.type_map.items():
        if (
            isinstance(graphql_type, GraphQLObjectType)
            and type_name not in GRAPHQL_SPEC_TYPES
        ):
            convert_names_in_schema_object(name_converter, graphql_type, schema)
        if isinstance(graphql_type, GraphQLInputObjectType):
            convert_names_in_schema_input(name_converter, graphql_type, schema)


def convert_names_in_schema_object(
    name_converter: SchemaNameConverter,
    graphql_object: GraphQLObjectType,
    schema: GraphQLSchema,
) -> None:
    for field_name, field in graphql_object.fields.items():
        if field.args:
            convert_names_in_schema_args(
                name_converter, field, schema, graphql_object.name, field_name
            )

        if field.resolve or field_name.lower() == field_name:
            continue

        field.resolve = resolve_to(
            name_converter(field_name, schema, (graphql_object.name, field_name))
        )


def convert_names_in_schema_args(
    name_converter: SchemaNameConverter,
    graphql_field: GraphQLField,
    schema: GraphQLSchema,
    object_name: str,
    field_name: str,
) -> None:
    for arg_name, arg in graphql_field.args.items():
        if arg.out_name or arg_name.lower() == arg_name:
            continue

        arg.out_name = name_converter(
            arg_name, schema, (object_name, field_name, arg_name)
        )


def convert_names_in_schema_input(
    name_converter: SchemaNameConverter,
    graphql_input: GraphQLInputObjectType,
    schema: GraphQLSchema,
) -> None:
    for field_name, field in graphql_input.fields.items():
        if field.out_name or field_name.lower() == field_name:
            continue

        field.out_name = name_converter(
            field_name, schema, (graphql_input.name, field_name)
        )


def default_schema_name_converter(graphql_name: str, *_) -> str:
    return convert_camel_case_to_snake(graphql_name)



from abc import ABC, abstractmethod
from inspect import isawaitable
from logging import Logger, LoggerAdapter
from typing import Any, Optional, Type, Union

from graphql import DocumentNode, ExecutionContext, GraphQLSchema, MiddlewareManager
from starlette.types import Receive, Scope, Send

from ...explorer import Explorer
from ...format_error import format_error
from ...types import (
    ContextValue,
    ErrorFormatter,
    GraphQLResult,
    OnComplete,
    OnConnect,
    OnDisconnect,
    OnOperation,
    QueryParser,
    RootValue,
    ValidationRules,
)
from ...utils import context_value_one_arg_deprecated


class GraphQLHandler(ABC):
    """Base class for ASGI connection handlers."""

    def __init__(self) -> None:
        """Initialize the handler instance with empty configuration."""
        self.schema: Optional[GraphQLSchema] = None
        self.context_value: Optional[ContextValue] = None
        self.debug: bool = False
        self.error_formatter: ErrorFormatter = format_error
        self.introspection: bool = True
        self.explorer: Optional[Explorer] = None
        self.logger: Union[None, str, Logger, LoggerAdapter] = None
        self.root_value: Optional[RootValue] = None
        self.query_parser: Optional[QueryParser] = None
        self.validation_rules: Optional[ValidationRules] = None
        self.execution_context_class: Optional[Type[ExecutionContext]] = None
        self.middleware_manager_class: Optional[Type[MiddlewareManager]] = None

    @abstractmethod
    async def handle(self, scope: Scope, receive: Receive, send: Send):
        """An entrypoint for the ASGI connection handler.

        This method is called by Ariadne ASGI GraphQL application. Subclasses
        should replace it with custom implementation.

        # Required arguments

        `scope`: The connection scope information, a dictionary that contains
        at least a type key specifying the protocol that is incoming.

        `receive`: an awaitable callable that will yield a new event dictionary
        when one is available.

        `send`: an awaitable callable taking a single event dictionary as a
        positional argument that will return once the send has been completed
        or the connection has been closed.

        Details about the arguments and their usage are described in the
        ASGI specification:

        https://asgi.readthedocs.io/en/latest/specs/main.html
        """
        raise NotImplementedError(
            "Subclasses of GraphQLHandler must implement the 'handle' method."
        )

    def configure(
        self,
        schema: GraphQLSchema,
        context_value: Optional[ContextValue] = None,
        root_value: Optional[RootValue] = None,
        query_parser: Optional[QueryParser] = None,
        validation_rules: Optional[ValidationRules] = None,
        debug: bool = False,
        introspection: bool = True,
        explorer: Optional[Explorer] = None,
        logger: Union[None, str, Logger, LoggerAdapter] = None,
        error_formatter: ErrorFormatter = format_error,
        execution_context_class: Optional[Type[ExecutionContext]] = None,
    ):
        """Configures the handler with options from the ASGI application.

        Called by Ariadne ASGI GraphQL application as part of its
        initialization, propagating the configuration to it's handlers.
        """
        self.context_value = context_value
        self.debug = debug
        self.error_formatter = error_formatter
        self.execution_context_class = execution_context_class
        self.introspection = introspection
        self.explorer = explorer
        self.logger = logger
        self.query_parser = query_parser
        self.root_value = root_value
        self.schema = schema
        self.validation_rules = validation_rules

    async def get_context_for_request(
        self,
        request: Any,
        data: dict,
    ) -> Any:
        """Returns GraphQL context value for ASGI connection.

        Resolves final context value from the `ContextValue` value passed to
        `context_value` option. If `context_value` is None, sets default context
        value instead, which is a `dict` with single `request` key that contains
        either `starlette.requests.Request` instance or
        `starlette.websockets.WebSocket` instance.

        # Required arguments

        `request`: an instance of ASGI connection. It's type depends on handler.

        `data`: a GraphQL data from connection.
        """
        if callable(self.context_value):
            try:
                context = self.context_value(request, data)  # type: ignore
            except TypeError:  # TODO: remove in 0.20
                context_value_one_arg_deprecated()
                context = self.context_value(request)  # type: ignore

            if isawaitable(context):
                context = await context
            return context

        return self.context_value or {"request": request}


class GraphQLHttpHandlerBase(GraphQLHandler):
    """Base class for ASGI HTTP connection handlers."""

    @abstractmethod
    async def handle_request(self, request: Any) -> Any:
        """Abstract method for handling the request.

        Should return valid ASGI response.
        """

    @abstractmethod
    async def execute_graphql_query(
        self,
        request: Any,
        data: Any,
        *,
        # Fast path for scenarios where context was already resolved
        # or query was already parsed
        context_value: Optional[Any] = None,
        query_document: Optional[DocumentNode] = None,
    ) -> GraphQLResult:
        """Abstract method for GraphQL query execution."""


class GraphQLWebsocketHandler(GraphQLHandler):
    """Base class for ASGI websocket connection handlers."""

    def __init__(
        self,
        on_connect: Optional[OnConnect] = None,
        on_disconnect: Optional[OnDisconnect] = None,
        on_operation: Optional[OnOperation] = None,
        on_complete: Optional[OnComplete] = None,
    ) -> None:
        """Initialize websocket handler with optional options specific to it.

        # Optional arguments:

        `on_connect`: an `OnConnect` callback used on new websocket connection.

        `on_disconnect`: an `OnDisconnect` callback used when existing
        websocket connection is closed.

        `on_operation`: an `OnOperation` callback, used when new GraphQL
        operation is received from websocket connection.

        `on_complete`: an `OnComplete` callback, used when GraphQL operation
        received over the websocket connection was completed.
        """
        super().__init__()
        self.http_handler: Optional[GraphQLHttpHandlerBase] = None

        self.on_connect: Optional[OnConnect] = on_connect
        self.on_disconnect: Optional[OnDisconnect] = on_disconnect
        self.on_operation: Optional[OnOperation] = on_operation
        self.on_complete: Optional[OnComplete] = on_complete

    @abstractmethod
    async def handle_websocket(self, websocket: Any):
        """Abstract method for handling the websocket connection."""

    def configure(
        self,
        *args,
        http_handler: Optional[GraphQLHttpHandlerBase] = None,
        **kwargs,
    ):
        """Configures the handler with options from the ASGI application.

        Called by Ariadne ASGI GraphQL application as part of its
        initialization, propagating the configuration to it's handlers.

        # Optional arguments

        `http_handler`: the `GraphQLHttpHandlerBase` subclass instance to use
        to execute the `Query` and `Mutation` operations made over the
        websocket connections.
        """
        super().configure(*args, **kwargs)
        self.http_handler = http_handler



from typing import Callable, Dict

from graphql.type import GraphQLSchema

from .objects import ObjectType
from .types import Subscriber


class SubscriptionType(ObjectType):
    """Bindable populating the Subscription type in a GraphQL schema with Python logic.

    Extends `ObjectType`, providing `source` decorator and `set_source` method, used
    to set subscription sources for it's fields.

    # Subscription sources ("subscribers")

    Subscription source is a function that is an async generator. This function is used
    to subscribe to source of events or messages. It can also filter the messages
    by not yielding them.

    Its signature is same as resolver:

    ```python
    async def source_fn(
        root_value: Any, info: GraphQLResolveInfo, **field_args
    ) -> Any:
        yield ...
    ```

    # Subscription resolvers

    Subscription resolvers are called with message returned from the source. Their role
    is to convert this message into Python representation of a type associated with
    subscription's field in GraphQL schema. Its called with message yielded from
    source function as first argument.

    ```python
    def resolver_fn(
        message: Any, info: GraphQLResolveInfo, **field_args
    ) -> Any:
        # Subscription resolver can be sync and async.
        return ...
    ```

    # GraphQL arguments

    When subscription field has arguments those arguments values are passed
    to both source and resolver functions.

    # Example source and resolver

    ```python
    from ariadne import SubscriptionType, make_executable_schema
    from broadcast import broadcast

    from .models import Post


    subscription_type = SubscriptionType()


    @subscription_type.source("post")
    async def source_post(*_, category: Optional[str] = None) -> dict:
        async with broadcast.subscribe(channel="NEW_POSTS") as subscriber:
            async for event in subscriber:
                message = json.loads(event.message)
                # Send message to resolver if we don't filter
                if not category or message["category"] == category:
                    yield message


    @subscription_type.field("post")
    async def resolve_post(
        message: dict, *_, category: Optional[str] = None
    ) -> Post:
        # Convert message to Post object that resolvers for Post type in
        # GraphQL schema understand.
        return await Post.get_one(id=message["post_id"])


    schema = make_executable_schema(
        \"\"\"
        type Query {
            \"Valid schema must define the Query type\"
            none: Int
        }

        type Subscription {
            post(category: ID): Post!
        }

        type Post {
            id: ID!
            author: String!
            text: String!
        }
        \"\"\",
        subscription_type
    )
    ```

    # Example chat

    [Ariadne GraphQL Chat Example](https://github.com/mirumee/ariadne-graphql-chat-example)
    is the Github repository with GraphQL chat application, using Redis message backend,
    Broadcaster library for publishing and subscribing to messages and React.js client
    using Apollo-Client subscriptions.
    """

    _subscribers: Dict[str, Subscriber]

    def __init__(self) -> None:
        """Initializes the `SubscriptionType` with a GraphQL name set to `Subscription`."""
        super().__init__("Subscription")
        self._subscribers = {}

    def source(self, name: str) -> Callable[[Subscriber], Subscriber]:
        """Return a decorator that sets decorated function as a source for named field.

        Wrapper for `create_register_subscriber` that on runtime validates `name` to be a
        string.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        bind decorated source to.
        """
        if not isinstance(name, str):
            raise ValueError(
                'source decorator should be passed a field name: @foo.source("name")'
            )
        return self.create_register_subscriber(name)

    def create_register_subscriber(
        self, name: str
    ) -> Callable[[Subscriber], Subscriber]:
        """Return a decorator that sets decorated function as a source for named field.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        bind decorated source to.
        """

        def register_subscriber(generator: Subscriber) -> Subscriber:
            self._subscribers[name] = generator
            return generator

        return register_subscriber

    def set_source(self, name, generator: Subscriber) -> Subscriber:
        """Set a source for the field name.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        set this source for.

        `generator`: a `Subscriber` function to use as an source.
        """
        self._subscribers[name] = generator
        return generator

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Binds this `SubscriptionType` instance to the instance of GraphQL schema.

        If it has any previously set subscription resolvers or source functions,
        those will be replaced with new ones from this instance.
        """
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)
        self.bind_resolvers_to_graphql_type(graphql_type)
        self.bind_subscribers_to_graphql_type(graphql_type)

    def bind_subscribers_to_graphql_type(self, graphql_type):
        """Binds this `SubscriptionType` instance's source functions.

        Source functions are set to fields `subscribe` attributes.
        """
        for field, subscriber in self._subscribers.items():
            if field not in graphql_type.fields:
                raise ValueError(
                    "Field %s is not defined on type %s" % (field, self.name)
                )

            graphql_type.fields[field].subscribe = subscriber



from ..types import (
    Extensions,
    MiddlewareList,
    Middlewares,
    OnComplete,
    OnConnect,
    OnDisconnect,
    OnOperation,
    Operation,
    WebSocketConnectionError,
)
from .graphql import GraphQL

__all__ = [
    "Extensions",
    "GraphQL",
    "MiddlewareList",
    "Middlewares",
    "OnComplete",
    "OnConnect",
    "OnDisconnect",
    "OnOperation",
    "Operation",
    "WebSocketConnectionError",
]



import logging
from typing import Union

from .utils import unwrap_graphql_error


def log_error(
    error: Exception,
    logger: Union[None, str, logging.Logger, logging.LoggerAdapter],
):
    original_error = unwrap_graphql_error(error)
    if original_error and original_error is not error:
        error.__suppress_context__ = True
        error.__cause__ = original_error

    if not logger:
        logger = "ariadne"
    if isinstance(logger, str):
        logger = logging.getLogger(logger)

    logger.error(error, exc_info=error)



from collections.abc import Mapping
from typing import Any, Optional

from graphql import default_field_resolver
from graphql.type import (
    GraphQLField,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
)

from .types import Resolver, SchemaBindable
from .utils import convert_camel_case_to_snake


class FallbackResolversSetter(SchemaBindable):
    """Bindable that recursively scans GraphQL schema for fields and explicitly
    sets their resolver to `graphql.default_field_resolver` package if
    they don't have any resolver set yet.

    > **Deprecated:** This class doesn't provide any utility for developers and
    only serves as a base for `SnakeCaseFallbackResolversSetter` which is being
    replaced by what we believe to be a better solution.
    >
    > Because of this we are deprecating this utility. It will be removed in future
    Ariadne release.
    """

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Scans GraphQL schema for types with fields that don't have set resolver."""
        for type_object in schema.type_map.values():
            if isinstance(type_object, GraphQLObjectType):
                self.add_resolvers_to_object_fields(type_object)

    def add_resolvers_to_object_fields(self, type_object: GraphQLObjectType) -> None:
        """Sets explicit default resolver on a fields of an object that don't have any."""
        for field_name, field_object in type_object.fields.items():
            self.add_resolver_to_field(field_name, field_object)

    def add_resolver_to_field(self, _: str, field_object: GraphQLField) -> None:
        """Sets `default_field_resolver` as a resolver on a field that doesn't have any."""
        if field_object.resolve is None:
            field_object.resolve = default_field_resolver


class SnakeCaseFallbackResolversSetter(FallbackResolversSetter):
    """Subclass of `FallbackResolversSetter` that uses case-converting resolver
    instead of `default_field_resolver`.

    > **Deprecated:** Use `convert_names_case` from `make_executable_schema`
    instead.
    """

    def add_resolver_to_field(
        self, field_name: str, field_object: GraphQLField
    ) -> None:
        """Sets case converting resolver on a field that doesn't have any."""
        if field_object.resolve is None:
            field_name = convert_camel_case_to_snake(field_name)
            field_object.resolve = resolve_to(field_name)


"""
Bindable instance of `FallbackResolversSetter`.

> **Deprecated:** This utility will be removed in future Ariadne release.
> 
> See `FallbackResolversSetter` for details.
"""
fallback_resolvers = FallbackResolversSetter()

"""
Bindable instance of `SnakeCaseFallbackResolversSetter`.

> **Deprecated:** Use `convert_names_case` from `make_executable_schema` 
instead.
"""
snake_case_fallback_resolvers = SnakeCaseFallbackResolversSetter()


def resolve_parent_field(parent: Any, field_name: str) -> Any:
    if isinstance(parent, Mapping):
        return parent.get(field_name)
    return getattr(parent, field_name, None)


def resolve_to(attr_name: str) -> Resolver:
    """Create a resolver that resolves to given attribute or dict key.

    Returns a resolver function that can be used as resolver.

    Usually not used directly  but through higher level features like aliases
    or schema names conversion.

    # Required arguments

    `attr_name`: a `str` with name of attribute or `dict` key to return from
    resolved object.
    """

    def resolver(parent: Any, info: GraphQLResolveInfo, **kwargs) -> Any:
        value = resolve_parent_field(parent, attr_name)
        if callable(value):
            return value(info, **kwargs)
        return value

    # pylint: disable=protected-access
    resolver._ariadne_alias_resolver = True  # type: ignore
    return resolver


def is_default_resolver(resolver: Optional[Resolver]) -> bool:
    """Test if resolver function is default resolver implemented by
    `graphql-core` or Ariadne.

    Returns `True` if resolver function is `None`, `graphql.default_field_resolver`
    or was created by Ariadne's `resolve_to` utility. Returns `False` otherwise.

    `True` is returned for `None` because query executor defaults to the
    `graphql.default_field_resolver` is there's no resolver function set on a
    field.

    # Required arguments

    `resolver`: an function `None` to test or `None`.
    """
    # pylint: disable=comparison-with-callable
    if not resolver or resolver == default_field_resolver:
        return True
    return hasattr(resolver, "_ariadne_alias_resolver")



import json
from inspect import isawaitable
from typing import Any, Optional, Type, cast

from graphql import DocumentNode, MiddlewareManager
from starlette.datastructures import UploadFile
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse, PlainTextResponse, Response
from starlette.types import Receive, Scope, Send

from ...explorer import Explorer
from ...constants import (
    DATA_TYPE_JSON,
    DATA_TYPE_MULTIPART,
)
from ...exceptions import HttpBadRequestError, HttpError
from ...file_uploads import combine_multipart_data
from ...graphql import graphql
from ...types import (
    ContextValue,
    ExtensionList,
    Extensions,
    GraphQLResult,
    MiddlewareList,
    Middlewares,
)
from .base import GraphQLHttpHandlerBase


class GraphQLHTTPHandler(GraphQLHttpHandlerBase):
    """Default ASGI handler for HTTP requests.

    Supports the `Query` and `Mutation` operations.
    """

    def __init__(
        self,
        extensions: Optional[Extensions] = None,
        middleware: Optional[Middlewares] = None,
        middleware_manager_class: Optional[Type[MiddlewareManager]] = None,
    ) -> None:
        """Initializes the HTTP handler.

        # Optional arguments

        `extensions`: an `Extensions` list or callable returning a
        list of extensions server should use during query execution. Defaults
        to no extensions.

        `middleware`: a `Middlewares` list or callable returning a list of
        middlewares server should use during query execution. Defaults to no
        middlewares.

        `middleware_manager_class`: a `MiddlewareManager` type or subclass to
        use for combining provided middlewares into single wrapper for resolvers
        by the server. Defaults to `graphql.MiddlewareManager`. Is only used
        if `extensions` or `middleware` options are set.
        """
        super().__init__()

        self.extensions = extensions
        self.middleware = middleware
        self.middleware_manager_class = middleware_manager_class or MiddlewareManager

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        """An entrypoint for the GraphQL HTTP handler.

        This method is called by the Ariadne ASGI GraphQL application to execute
        queries done using the HTTP protocol.

        It creates the `starlette.requests.Request` instance, calls
        `handle_request` method with it, then sends response back to the client.

        # Required arguments

        `scope`: The connection scope information, a dictionary that contains
        at least a type key specifying the protocol that is incoming.

        `receive`: an awaitable callable that will yield a new event dictionary
        when one is available.

        `send`: an awaitable callable taking a single event dictionary as a
        positional argument that will return once the send has been completed
        or the connection has been closed.

        Details about the arguments and their usage are described in the
        ASGI specification:

        https://asgi.readthedocs.io/en/latest/specs/main.html
        """
        request = Request(scope=scope, receive=receive)
        response = await self.handle_request(request)
        await response(scope, receive, send)

    async def handle_request(self, request: Request) -> Response:
        """Handle GraphQL request and return response for the client.

        Is called by the `handle` method and `handle_request` method of the
        ASGI GraphQL application.

        Handles three HTTP methods:

        `GET`: returns GraphQL explorer or 405 error response if explorer or
        introspection is disabled.

        `POST`: executes the GraphQL query from either `application/json` or
        `multipart/form-data` requests.

        `OPTIONS`: returns supported HTTP methods.

        Returns Starlette's `Response` instance, which is also works in FastAPI.

        # Required arguments:

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        if request.method == "GET" and self.introspection and self.explorer:
            # only render explorer when introspection is enabled
            return await self.render_explorer(request, self.explorer)

        if request.method == "POST":
            return await self.graphql_http_server(request)

        return self.handle_not_allowed_method(request)

    async def render_explorer(self, request: Request, explorer: Explorer) -> Response:
        """Return a HTML response with GraphQL explorer.

        # Required arguments:

        `request`: the `Request` instance from Starlette or FastAPI.

        `explorer`: an `Explorer` instance that implements the
        `html(request: Request)` method which returns either the `str` with HTML
        or `None`. If explorer returns `None`, `405` method not allowed response
        is returned instead.
        """
        explorer_html = explorer.html(request)
        if isawaitable(explorer_html):
            explorer_html = await explorer_html
        if explorer_html:
            return HTMLResponse(explorer_html)

        return self.handle_not_allowed_method(request)

    async def graphql_http_server(self, request: Request) -> Response:
        """Handles the HTTP request with GraphQL query.

        Extracts GraphQL query data from requests and then executes it using
        the `execute_graphql_query` method.

        Returns the JSON response from Sta

        If request's data was invalid or missing, plaintext response with
        error message and 400 status code is returned instead.

        # Required arguments:

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        try:
            data = await self.extract_data_from_request(request)
        except HttpError as error:
            return PlainTextResponse(error.message or error.status, status_code=400)

        success, result = await self.execute_graphql_query(request, data)
        return await self.create_json_response(request, result, success)

    async def extract_data_from_request(self, request: Request):
        """Extracts GraphQL request data from request.

        Returns a `dict` with GraphQL query data that was not yet validated.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        content_type = request.headers.get("Content-Type", "")
        content_type = content_type.split(";")[0]

        if content_type == DATA_TYPE_JSON:
            return await self.extract_data_from_json_request(request)
        if content_type == DATA_TYPE_MULTIPART:
            return await self.extract_data_from_multipart_request(request)

        raise HttpBadRequestError(
            "Posted content must be of type {} or {}".format(
                DATA_TYPE_JSON, DATA_TYPE_MULTIPART
            )
        )

    async def extract_data_from_json_request(self, request: Request):
        """Extracts GraphQL data from JSON request.

        Returns a `dict` with GraphQL query data that was not yet validated.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        try:
            return await request.json()
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError("Request body is not a valid JSON") from ex

    async def extract_data_from_multipart_request(self, request: Request):
        """Extracts GraphQL data from `multipart/form-data` request.

        Returns an unvalidated `dict` with GraphQL query data.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        try:
            request_body = await request.form()
        except ValueError as ex:
            raise HttpBadRequestError(
                "Request body is not a valid multipart/form-data"
            ) from ex

        try:
            operations = json.loads(cast(Any, request_body.get("operations")))
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError(
                "Request 'operations' multipart field is not a valid JSON"
            ) from ex
        try:
            files_map = json.loads(cast(Any, request_body.get("map")))
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError(
                "Request 'map' multipart field is not a valid JSON"
            ) from ex

        request_files = {
            key: value
            for key, value in request_body.items()
            if isinstance(value, UploadFile)
        }

        return combine_multipart_data(operations, files_map, request_files)

    async def execute_graphql_query(
        self,
        request: Any,
        data: Any,
        *,
        context_value: Any = None,
        query_document: Optional[DocumentNode] = None,
    ) -> GraphQLResult:
        """Executes GraphQL query from `request` and returns `GraphQLResult`.

        Creates GraphQL `ContextValue`, initializes extensions and middlewares,
        then runs the `graphql` function from Ariadne to execute the query.

        # Requires arguments

        `request`: the `Request` instance from Starlette or FastAPI.

        `data`: a GraphQL data.

        # Optional arguments

        `context_value`: a `ContextValue` for this request.

        `query_document`: an already parsed GraphQL query. Setting this option
        will prevent `graphql` from parsing `query` string from `data` second time.
        """
        if context_value is None:
            context_value = await self.get_context_for_request(request, data)

        extensions = await self.get_extensions_for_request(request, context_value)
        middleware = await self.get_middleware_for_request(request, context_value)

        if self.schema is None:
            raise TypeError("schema is not set, call configure method to initialize it")

        return await graphql(
            self.schema,
            data,
            context_value=context_value,
            root_value=self.root_value,
            query_parser=self.query_parser,
            query_document=query_document,
            validation_rules=self.validation_rules,
            debug=self.debug,
            introspection=self.introspection,
            logger=self.logger,
            error_formatter=self.error_formatter,
            extensions=extensions,
            middleware=middleware,
            middleware_manager_class=self.middleware_manager_class,
            execution_context_class=self.execution_context_class,
        )

    async def get_extensions_for_request(
        self, request: Any, context: Optional[ContextValue]
    ) -> ExtensionList:
        """Returns extensions to use when handling the GraphQL request.

        Returns `ExtensionList`, a list of extensions to use or `None`.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.

        `context`: a `ContextValue` for this request.
        """
        if callable(self.extensions):
            extensions = self.extensions(request, context)
            if isawaitable(extensions):
                extensions = await extensions  # type: ignore
            return extensions
        return self.extensions

    async def get_middleware_for_request(
        self, request: Any, context: Optional[ContextValue]
    ) -> MiddlewareList:
        """Returns GraphQL middlewares to use when handling the GraphQL request.

        Returns `MiddlewareList`, a list of middlewares to use or `None`.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.

        `context`: a `ContextValue` for this request.
        """
        middleware = self.middleware
        if callable(middleware):
            middleware = middleware(request, context)
            if isawaitable(middleware):
                middleware = await middleware  # type: ignore
        if middleware:
            return cast(MiddlewareList, middleware)
        return None

    async def create_json_response(
        self,
        request: Request,  # pylint: disable=unused-argument
        result: dict,
        success: bool,
    ) -> Response:
        """Creates JSON response from GraphQL's query result.

        Returns Starlette's `JSONResponse` instance that's also compatible
        with FastAPI. If `success` is `True`, response's status code is 200.
        Status code 400 is used otherwise.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.

        `result`: a JSON-serializable `dict` with query result.

        `success`: a `bool` specifying if
        """
        status_code = 200 if success else 400
        return JSONResponse(result, status_code=status_code)

    def handle_not_allowed_method(self, request: Request):
        """Handles request for unsupported HTTP method.

        Returns 200 response for `OPTIONS` request and 405 response for other
        methods. All responses have empty body.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        allowed_methods = ["OPTIONS", "POST"]
        if self.introspection:
            allowed_methods.append("GET")
        allow_header = {"Allow": ", ".join(allowed_methods)}

        if request.method == "OPTIONS":
            return Response(headers=allow_header)

        return Response(status_code=405, headers=allow_header)



DATA_TYPE_JSON = "application/json"
DATA_TYPE_MULTIPART = "multipart/form-data"

CONTENT_TYPE_JSON = "application/json; charset=UTF-8"
CONTENT_TYPE_TEXT_HTML = "text/html; charset=UTF-8"
CONTENT_TYPE_TEXT_PLAIN = "text/plain; charset=UTF-8"

HTTP_STATUS_200_OK = "200 OK"
HTTP_STATUS_400_BAD_REQUEST = "400 Bad Request"
HTTP_STATUS_405_METHOD_NOT_ALLOWED = "405 Method Not Allowed"



#
# The code herein implements (ports) the apollo-federation specs.
# Special thanks go to the apollo team for their great work!
# https://www.apollographql.com/docs/apollo-server/federation/federation-spec/
#

from .interfaces import FederatedInterfaceType
from .objects import FederatedObjectType
from .schema import make_federated_schema



import json
from typing import Dict, Optional, Union

from .explorer import Explorer
from .template import read_template, render_template

PLAYGROUND_HTML = read_template("playground.html")

SettingsDict = Dict[str, Union[str, int, bool, Dict[str, str]]]


class ExplorerPlayground(Explorer):
    def __init__(
        self,
        title: str = "Ariadne GraphQL",
        editor_cursor_shape: Optional[str] = None,
        editor_font_family: Optional[str] = None,
        editor_font_size: Optional[int] = None,
        editor_reuse_headers: Optional[bool] = None,
        editor_theme: Optional[str] = None,
        general_beta_updates: Optional[bool] = None,
        prettier_print_width: Optional[int] = None,
        prettier_tab_width: Optional[int] = None,
        prettier_use_tabs: Optional[bool] = None,
        request_credentials: Optional[str] = None,
        request_global_headers: Optional[Dict[str, str]] = None,
        schema_polling_enable: Optional[bool] = None,
        schema_polling_endpoint_filter: Optional[str] = None,
        schema_polling_interval: Optional[int] = None,
        schema_disable_comments: Optional[bool] = None,
        tracing_hide_tracing_response: Optional[bool] = None,
        tracing_tracing_supported: Optional[bool] = None,
        query_plan_hide_query_plan_response: Optional[bool] = None,
    ) -> None:
        settings = self.build_settings(
            editor_cursor_shape=editor_cursor_shape,
            editor_font_family=editor_font_family,
            editor_font_size=editor_font_size,
            editor_reuse_headers=editor_reuse_headers,
            editor_theme=editor_theme,
            general_beta_updates=general_beta_updates,
            prettier_print_width=prettier_print_width,
            prettier_tab_width=prettier_tab_width,
            prettier_use_tabs=prettier_use_tabs,
            request_credentials=request_credentials,
            request_global_headers=request_global_headers,
            schema_polling_enable=schema_polling_enable,
            schema_polling_endpoint_filter=schema_polling_endpoint_filter,
            schema_polling_interval=schema_polling_interval,
            schema_disable_comments=schema_disable_comments,
            tracing_hide_tracing_response=tracing_hide_tracing_response,
            tracing_tracing_supported=tracing_tracing_supported,
            query_plan_hide_query_plan_response=query_plan_hide_query_plan_response,
        )

        self.parsed_html = render_template(
            PLAYGROUND_HTML,
            {
                "title": title,
                "settings": json.dumps(settings) if settings else None,
            },
        )

    def build_settings(
        self,
        editor_cursor_shape: Optional[str] = None,
        editor_font_family: Optional[str] = None,
        editor_font_size: Optional[int] = None,
        editor_reuse_headers: Optional[bool] = None,
        editor_theme: Optional[str] = None,
        general_beta_updates: Optional[bool] = None,
        prettier_print_width: Optional[int] = None,
        prettier_tab_width: Optional[int] = None,
        prettier_use_tabs: Optional[bool] = None,
        request_credentials: Optional[str] = None,
        request_global_headers: Optional[Dict[str, str]] = None,
        schema_polling_enable: Optional[bool] = None,
        schema_polling_endpoint_filter: Optional[str] = None,
        schema_polling_interval: Optional[int] = None,
        schema_disable_comments: Optional[bool] = None,
        tracing_hide_tracing_response: Optional[bool] = None,
        tracing_tracing_supported: Optional[bool] = None,
        query_plan_hide_query_plan_response: Optional[bool] = None,
    ) -> SettingsDict:
        settings: SettingsDict = {}
        if editor_cursor_shape:
            settings["editor.cursorShape"] = editor_cursor_shape
        if editor_font_family:
            settings["editor.fontFamily"] = editor_font_family
        if editor_font_size:
            settings["editor.fontSize"] = editor_font_size
        if editor_reuse_headers is not None:
            settings["editor.reuseHeaders"] = editor_reuse_headers
        if editor_theme:
            settings["editor.theme"] = editor_theme
        if general_beta_updates is not None:
            settings["general.betaUpdates"] = general_beta_updates
        if prettier_print_width:
            settings["prettier.printWidth"] = prettier_print_width
        if prettier_tab_width:
            settings["prettier.tabWidth"] = prettier_tab_width
        if prettier_use_tabs is not None:
            settings["prettier.useTabs"] = prettier_use_tabs
        if request_credentials:
            settings["request.credentials"] = request_credentials
        if request_global_headers:
            settings["request.globalHeaders"] = request_global_headers
        if schema_polling_enable is not None:
            settings["schema.polling.enable"] = schema_polling_enable
        if schema_polling_endpoint_filter:
            settings["schema.polling.endpointFilter"] = schema_polling_endpoint_filter
        if schema_polling_interval:
            settings["schema.polling.interval"] = schema_polling_interval
        if schema_disable_comments is not None:
            settings["schema.disableComments"] = schema_disable_comments
        if tracing_hide_tracing_response is not None:
            settings["tracing.hideTracingResponse"] = tracing_hide_tracing_response
        if tracing_tracing_supported is not None:
            settings["tracing.tracingSupported"] = tracing_tracing_supported
        if query_plan_hide_query_plan_response is not None:
            settings[
                "queryPlan.hideQueryPlanResponse"
            ] = query_plan_hide_query_plan_response

        return settings

    def html(self, _):
        return self.parsed_html



import asyncio
from contextlib import suppress
from datetime import timedelta
from inspect import isawaitable
from typing import Any, AsyncGenerator, Dict, List, Optional, cast

from graphql import GraphQLError
from graphql.language import OperationType
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState

from ...graphql import subscribe, parse_query, validate_data
from ...logger import log_error
from ...types import (
    ExecutionResult,
    Operation,
)
from ...utils import get_operation_type
from .base import GraphQLWebsocketHandler


class ClientContext:
    def __init__(self) -> None:
        self.connection_acknowledged: bool = False
        self.connection_init_timeout_task: Optional[asyncio.Task] = None
        self.connection_init_received: bool = False
        self.operations: Dict[str, Operation] = {}
        self.operation_tasks: Dict[str, asyncio.Task] = {}
        self.websocket: WebSocket


class GraphQLTransportWSHandler(GraphQLWebsocketHandler):
    """Implementation of the (newer) graphql-transport-ws subprotocol
    from the graphql-ws library.

    For more details see it's GH page:

    https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md
    """

    GQL_CONNECTION_INIT = "connection_init"  # Client -> Server
    GQL_CONNECTION_ACK = "connection_ack"  # Server -> Client
    GQL_PING = "ping"  # Client -> Server, Server -> Client
    GQL_PONG = "pong"  # Client -> Server, Server -> Client
    GQL_SUBSCRIBE = "subscribe"  # Client -> Server
    GQL_NEXT = "next"  # Server -> Client
    GQL_ERROR = "error"  # Server -> Client
    GQL_COMPLETE = "complete"  # Client -> Server, Server -> Client

    def __init__(
        self,
        *args,
        connection_init_wait_timeout: timedelta = timedelta(minutes=1),
        **kwargs,
    ) -> None:
        """Initializes the websocket handler.

        # Optional arguments

        `connection_init_wait_timeout`: a `timedelta` with timeout for new
        websocket connections before first message is received. Defaults to
        60 seconds.
        """
        super().__init__(*args, **kwargs)

        self.connection_init_wait_timeout = connection_init_wait_timeout

    async def handle(self, scope: Scope, receive: Receive, send: Send):
        """An entrypoint for the GraphQL WebSocket handler.

        This method is called by the Ariadne ASGI GraphQL application to handle
        the websocket connections.

        It creates the `starlette.websockets.WebSocket` instance and calls
        `handle_websocket` method with it.

        # Required arguments

        `scope`: The connection scope information, a dictionary that contains
        at least a type key specifying the protocol that is incoming.

        `receive`: an awaitable callable that will yield a new event dictionary
        when one is available.

        `send`: an awaitable callable taking a single event dictionary as a
        positional argument that will return once the send has been completed
        or the connection has been closed.

        Details about the arguments and their usage are described in the
        ASGI specification:

        https://asgi.readthedocs.io/en/latest/specs/main.html
        """
        websocket = WebSocket(scope=scope, receive=receive, send=send)
        await self.handle_websocket(websocket)

    async def handle_websocket(self, websocket: WebSocket):
        """Handle GraphQL the WebSocket connection.

        Is called by the `handle` method and `handle_websocket` method of the
        ASGI GraphQL application.

        # Required arguments:

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.
        """
        client_context = ClientContext()
        timeout_handler = self.handle_connection_init_timeout(websocket, client_context)
        client_context.connection_init_timeout_task = asyncio.create_task(
            timeout_handler
        )

        await websocket.accept("graphql-transport-ws")
        try:
            while WebSocketState.DISCONNECTED not in (
                websocket.client_state,
                websocket.application_state,
            ):
                message = await websocket.receive_json()
                await self.handle_websocket_message(websocket, message, client_context)
        except WebSocketDisconnect:
            pass
        finally:
            for operation_id in list(client_context.operations.keys()):
                await self.stop_websocket_operation(
                    websocket, operation_id, client_context
                )

            try:
                if self.on_disconnect:
                    result = self.on_disconnect(websocket)
                    if result and isawaitable(result):
                        await result
            except Exception as error:
                if not isinstance(error, GraphQLError):
                    error = GraphQLError(str(error), original_error=error)
                log_error(error, self.logger)

    async def handle_connection_init_timeout(
        self, websocket: WebSocket, client_context: ClientContext
    ):
        delay = self.connection_init_wait_timeout.total_seconds()
        await asyncio.sleep(delay=delay)
        if client_context.connection_init_received:
            return
        if WebSocketState.DISCONNECTED not in (
            websocket.client_state,
            websocket.application_state,
        ):
            # 4408: Connection initialisation timeout
            await websocket.close(code=4408)

    async def handle_websocket_message(
        self,
        websocket: WebSocket,
        message: dict,
        client_context: ClientContext,
    ):
        """Handles new message from websocket connection.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `message`: a `dict` with message payload.

        `client_context`: a `ClientContext` object with extra state of current
        websocket connection.
        """
        operation_id = cast(str, message.get("id"))
        message_type = cast(str, message.get("type"))

        if message_type == GraphQLTransportWSHandler.GQL_CONNECTION_INIT:
            await self.handle_websocket_connection_init_message(
                websocket, message, client_context
            )
        elif message_type == GraphQLTransportWSHandler.GQL_PING:
            await self.handle_websocket_ping_message(websocket, client_context)
        elif message_type == GraphQLTransportWSHandler.GQL_PONG:
            await self.handle_websocket_pong_message(websocket, client_context)
        elif message_type == GraphQLTransportWSHandler.GQL_COMPLETE:
            await self.handle_websocket_complete_message(
                websocket, operation_id, client_context
            )
        elif message_type == GraphQLTransportWSHandler.GQL_SUBSCRIBE:
            await self.handle_websocket_subscribe(
                websocket, message.get("payload"), operation_id, client_context
            )
        else:
            await self.handle_websocket_invalid_type(websocket)

    async def handle_websocket_connection_init_message(
        self,
        websocket: WebSocket,
        message: dict,
        client_context: ClientContext,
    ):
        """Handles `connection_init` websocket message.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `message`: a `dict` with message payload.

        `client_context`: a `ClientContext` object with extra state of current
        websocket connection.
        """
        if client_context.connection_init_received:
            # 4429: Too many initialisation requests
            await websocket.close(code=4429)
            return

        client_context.connection_init_received = True

        try:
            if self.on_connect:
                result = self.on_connect(websocket, message.get("payload"))
                if result and isawaitable(result):
                    await result

            await websocket.send_json(
                {"type": GraphQLTransportWSHandler.GQL_CONNECTION_ACK}
            )
            client_context.connection_acknowledged = True
        except Exception as error:
            log_error(error, self.logger)
            await websocket.close()

    async def handle_websocket_ping_message(
        self,
        websocket: WebSocket,
        client_context: ClientContext,  # pylint: disable=unused-argument
    ):
        """Handles `ping` websocket message, answering with `pong` message.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `client_context`: a `ClientContext` object with extra state of current
        websocket connection.
        """
        await websocket.send_json({"type": GraphQLTransportWSHandler.GQL_PONG})

    async def handle_websocket_pong_message(
        self,
        websocket: WebSocket,
        client_context: ClientContext,  # pylint: disable=unused-argument
    ):
        """Handles `pong` websocket message.

        Unlike `ping` message, `pong` is unidirectional heartbeat sent by the
        client to the server. It doesn't require a result.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `client_context`: a `ClientContext` object with extra state of current
        websocket connection.
        """

    async def handle_websocket_complete_message(
        self,
        websocket: WebSocket,
        operation_id: str,
        client_context: ClientContext,
    ):
        """Handles `complete` websocket message.

        `complete` message tells the GraphQL server to stop sending events for
        GraphQL operation specified in the message

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `operation_id`: a `str` with id of operation that should be stopped.

        `client_context`: a `ClientContext` object with extra state of current
        websocket connection.
        """
        await self.stop_websocket_operation(websocket, operation_id, client_context)

    async def handle_websocket_subscribe(
        self,
        websocket: WebSocket,
        data: Any,
        operation_id: str,
        client_context: ClientContext,
    ):
        """Handles `subscribe` websocket message.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `data`: any data from `subscribe` message.

        `operation_id`: a `str` with id of new subscribe operation.

        `client_context`: a `ClientContext` object with extra state of current
        websocket connection.
        """
        if not client_context.connection_acknowledged:
            await websocket.close(code=4401)
            return

        if operation_id in client_context.operations:
            await websocket.close(code=4409)
            return

        validate_data(data)

        context_value = await self.get_context_for_request(websocket, data)

        try:
            query_document = parse_query(context_value, self.query_parser, data)
            operation_type = get_operation_type(
                query_document, data.get("operationName")
            )
        except GraphQLError as error:
            log_error(error, self.logger)
            await websocket.send_json(
                {
                    "type": GraphQLTransportWSHandler.GQL_ERROR,
                    "id": operation_id,
                    "payload": self.error_formatter(error, self.debug),
                }
            )
            return

        if operation_type == OperationType.SUBSCRIPTION:
            if self.schema is None:
                raise TypeError(
                    "schema is not set, call configure method to initialize it"
                )

            success, results_producer = await subscribe(
                self.schema,
                data,
                context_value=context_value,
                root_value=self.root_value,
                query_document=query_document,
                validation_rules=self.validation_rules,
                debug=self.debug,
                introspection=self.introspection,
                logger=self.logger,
                error_formatter=self.error_formatter,
            )
        else:
            if self.http_handler is None:
                raise TypeError(
                    "http_handler is not set, call configure method to initialize it"
                )

            success, result = await self.http_handler.execute_graphql_query(
                websocket,
                data,
                context_value=context_value,
                query_document=query_document,
            )

            async def get_results():
                yield result

            # if success then AsyncGenerator is expected, for error it will be List
            results_producer = get_results() if success else [result]

        if not success:
            results_producer = cast(List[dict], results_producer)
            await websocket.send_json(
                {
                    "type": GraphQLTransportWSHandler.GQL_ERROR,
                    "id": operation_id,
                    "payload": results_producer[0],
                }
            )
        else:
            results_producer = cast(
                AsyncGenerator[ExecutionResult, None], results_producer
            )
            client_context.operations[operation_id] = Operation(
                id=operation_id,
                name=data.get("operationName"),
                generator=results_producer,
            )

            if self.on_operation:
                try:
                    result = self.on_operation(
                        websocket, client_context.operations[operation_id]
                    )
                    if result and isawaitable(result):
                        await result
                except Exception as error:
                    if not isinstance(error, GraphQLError):
                        error = GraphQLError(str(error), original_error=error)
                    log_error(error, self.logger)

            # store Task in the operation_tasks list so that we can cancel such task
            # if client sends the "complete" message
            client_context.operation_tasks[operation_id] = asyncio.ensure_future(
                self.observe_async_results(
                    websocket, results_producer, operation_id, client_context
                )
            )

    async def handle_websocket_invalid_type(self, websocket: WebSocket):
        """Handles unsupported or invalid websocket message.

        Closes open websocket connection with error code `4400`.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.
        """
        await websocket.close(code=4400)

    async def handle_on_complete(
        self,
        websocket: WebSocket,
        operation: Operation,
    ):
        """Handles completed websocket operation.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `operation`: a completed `Operation`.
        """
        if self.on_complete:
            try:
                result = self.on_complete(websocket, operation)
                if result and isawaitable(result):
                    await result
            except Exception as error:
                if not isinstance(error, GraphQLError):
                    error = GraphQLError(str(error), original_error=error)
                log_error(error, self.logger)

    async def stop_websocket_operation(
        self,
        websocket: WebSocket,
        operation_id: str,
        client_context: ClientContext,
    ) -> None:
        """Stops specified GraphQL operation for given connection and context.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `operation_id`: a `str` with id of operation to stop.

        `client_context`: a `ClientContext` object with extra state of current
        websocket connection.
        """
        if operation_id not in client_context.operations:
            return

        operation = client_context.operations.pop(operation_id)
        await self.handle_on_complete(websocket, operation)

        otask = client_context.operation_tasks.pop(operation_id)
        otask.cancel()
        with suppress(asyncio.CancelledError):
            await otask
        await operation.generator.aclose()

    async def observe_async_results(
        self,
        websocket: WebSocket,
        results_producer: AsyncGenerator,
        operation_id: str,
        client_context: ClientContext,
    ) -> None:
        """Converts results from Ariadne's `subscribe` generator into websocket
        messages it next sends to the client.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `results_producer`: the `AsyncGenerator` returned from Ariadne's
        `subscribe` function.

        `operation_id`: a `str` with id of operation.

        `client_context`: a `ClientContext` object with extra state of current
        websocket connection.
        """
        try:
            async for result in results_producer:
                if not isinstance(result, dict):
                    payload = {}
                    if result.data:
                        payload["data"] = result.data
                    if result.errors:
                        for error in result.errors:
                            log_error(error, self.logger)
                        payload["errors"] = [
                            self.error_formatter(error, self.debug)
                            for error in result.errors
                        ]
                else:
                    payload = result

                await websocket.send_json(
                    {
                        "type": GraphQLTransportWSHandler.GQL_NEXT,
                        "id": operation_id,
                        "payload": payload,
                    }
                )
        except asyncio.CancelledError:  # pylint: disable=W0706
            # if asyncio Task is cancelled then CancelledError is thrown in the coroutine
            raise
        except Exception as error:
            if not isinstance(error, GraphQLError):
                error = GraphQLError(str(error), original_error=error)
            log_error(error, self.logger)
            payload = {"errors": [self.error_formatter(error, self.debug)]}

            await websocket.send_json(
                {
                    "type": GraphQLTransportWSHandler.GQL_NEXT,
                    "id": operation_id,
                    "payload": payload,
                }
            )

        operation = client_context.operations.pop(operation_id)
        del client_context.operation_tasks[operation_id]
        await self.handle_on_complete(websocket, operation)

        if WebSocketState.DISCONNECTED not in (
            websocket.client_state,
            websocket.application_state,
        ):
            await websocket.send_json(
                {"type": GraphQLTransportWSHandler.GQL_COMPLETE, "id": operation_id}
            )
